/*
 * Project: MoleCuilder
 * Description: creates and alters molecular systems
 * Copyright (C)  2010 University of Bonn. All rights reserved.
 * Please see the LICENSE file or "Copyright notice" in builder.cpp for details.
 */

/*
 * bondgraph.cpp
 *
 *  Created on: Oct 29, 2009
 *      Author: heber
 */

// include config.h
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "CodePatterns/MemDebug.hpp"

#include <iostream>

#include "atom.hpp"
#include "bond.hpp"
#include "bondgraph.hpp"
#include "element.hpp"
#include "CodePatterns/Info.hpp"
#include "CodePatterns/Verbose.hpp"
#include "CodePatterns/Log.hpp"
#include "molecule.hpp"
#include "parser.hpp"
#include "periodentafel.hpp"
#include "LinearAlgebra/Vector.hpp"

const double BondGraph::BondThreshold = 0.4;   //!< CSD threshold in bond check which is the width of the interval whose center is the sum of the covalent radii

/** Constructor of class BondGraph.
 * This classes contains typical bond lengths and thus may be used to construct a bond graph for a given molecule.
 */
BondGraph::BondGraph(bool IsA) :
    BondLengthMatrix(NULL),
    max_distance(0),
    IsAngstroem(IsA)
{};

/** Destructor of class BondGraph.
 */
BondGraph::~BondGraph()
{
  if (BondLengthMatrix != NULL) {
    delete(BondLengthMatrix);
  }
};

/** Parses the bond lengths in a given file and puts them int a matrix form.
 * Allocates \a MatrixContainer for BondGraph::BondLengthMatrix, using MatrixContainer::ParseMatrix(),
 * but only if parsing is successful. Otherwise variable is left as NULL.
 * \param &input input stream to parse table from
 * \return true - success in parsing file, false - failed to parse the file
 */
bool BondGraph::LoadBondLengthTable(std::istream &input)
{
  Info FunctionInfo(__func__);
  bool status = true;
  MatrixContainer *TempContainer = NULL;

  // allocate MatrixContainer
  if (BondLengthMatrix != NULL) {
    DoLog(1) && (Log() << Verbose(1) << "MatrixContainer for Bond length already present, removing." << endl);
    delete(BondLengthMatrix);
  }
  TempContainer = new MatrixContainer;

  // parse in matrix
  if ((input.good()) && (status = TempContainer->ParseMatrix(input, 0, 1, 0))) {
    DoLog(1) && (Log() << Verbose(1) << "Parsing bond length matrix successful." << endl);
  } else {
    DoeLog(1) && (eLog()<< Verbose(1) << "Parsing bond length matrix failed." << endl);
    status = false;
  }

  // find greatest distance
  max_distance=0;
  if (status) {
    for(int i=0;i<TempContainer->RowCounter[0];i++)
      for(int j=i;j<TempContainer->ColumnCounter[0];j++)
        if (TempContainer->Matrix[0][i][j] > max_distance)
          max_distance = TempContainer->Matrix[0][i][j];
  }
  max_distance += BondThreshold;

  if (status) // set to not NULL only if matrix was parsed
    BondLengthMatrix = TempContainer;
  else {
    BondLengthMatrix = NULL;
    delete(TempContainer);
  }
  return status;
};

/** Parses the bond lengths in a given file and puts them int a matrix form.
 * \param *out output stream for debugging
 * \param *mol molecule with atoms
 * \return true - success, false - failed to construct bond structure
 */
bool BondGraph::ConstructBondGraph(molecule * const mol)
{
  Info FunctionInfo(__func__);
  bool status = true;

  if (mol->empty()) // only construct if molecule is not empty
    return false;

  SetMaxDistanceToMaxOfCovalentRadii(mol);
  mol->CreateAdjacencyList(max_distance, IsAngstroem, &BondGraph::getMinMaxDistance, this);

  return status;
};

/** Returns the entry for a given index pair.
 * \param firstelement index/atom number of first element (row index)
 * \param secondelement index/atom number of second element (column index)
 * \note matrix is of course symmetric.
 */
double BondGraph::GetBondLength(int firstZ, int secondZ)
{
  std::cout << "Request for length between " << firstZ << " and " << secondZ << ": ";
  if (BondLengthMatrix == NULL) {
    std::cout << "-1." << std::endl;
    return( -1. );
  } else {
    std::cout << BondLengthMatrix->Matrix[0][firstZ][secondZ] << std::endl;
    return (BondLengthMatrix->Matrix[0][firstZ][secondZ]);
  }
};

/** Determines the maximum of all element::CovalentRadius for elements present in \a *mol.
 * \param *out output stream for debugging
 * \param *mol molecule with all atoms and their respective elements.
 */
double BondGraph::SetMaxDistanceToMaxOfCovalentRadii(const molecule * const mol)
{
  Info FunctionInfo(__func__);
  max_distance = 0.;

  for (molecule::const_iterator iter = mol->begin(); iter != mol->end(); ++iter) {
    if ((*iter)->getType()->getCovalentRadius() > max_distance)
      max_distance = (*iter)->getType()->getCovalentRadius();
  }
  max_distance *= 2.;
  max_distance += BondThreshold;

  return max_distance;
};

/** Returns the maximum distance (e.g. necessary for LinkedCell).
 * \return BondGraph::max_distance
 */
double BondGraph::getMaxDistance() const
{
  return max_distance;
}


/** Returns bond criterion for given pair based on covalent radius.
 * \param *Walker first BondedParticle
 * \param *OtherWalker second BondedParticle
 * \param &MinDistance lower bond bound on return
 * \param &MaxDistance upper bond bound on return
 * \param IsAngstroem whether units are in angstroem or bohr radii
 */
void BondGraph::CovalentMinMaxDistance(BondedParticle * const Walker, BondedParticle * const OtherWalker, double &MinDistance, double &MaxDistance, bool IsAngstroem)
{
  MinDistance = OtherWalker->getType()->getCovalentRadius() + Walker->getType()->getCovalentRadius();
  MinDistance *= (IsAngstroem) ? 1. : 1. / AtomicLengthToAngstroem;
  MaxDistance = MinDistance + BondThreshold;
  MinDistance -= BondThreshold;
};

/** Returns bond criterion for given pair based on a bond length matrix.
 * The matrix should be contained in \a this BondGraph and contain an element-
 * to-element length.
 * \param *Walker first BondedParticle
 * \param *OtherWalker second BondedParticle
 * \param &MinDistance lower bond bound on return
 * \param &MaxDistance upper bond bound on return
 * \param IsAngstroem whether units are in angstroem or bohr radii
 */
void BondGraph::BondLengthMatrixMinMaxDistance(BondedParticle * const Walker, BondedParticle * const OtherWalker, double &MinDistance, double &MaxDistance, bool IsAngstroem)
{
  ASSERT(BondLengthMatrix != NULL,
      "BondGraph::BondLengthMatrixMinMaxDistance() called without NULL BondLengthMatrix.");
  MinDistance = GetBondLength(Walker->getType()->getAtomicNumber()-1, OtherWalker->getType()->getAtomicNumber()-1);
  MinDistance *= (IsAngstroem) ? 1. : 1. / AtomicLengthToAngstroem;
  MaxDistance = MinDistance + BondThreshold;
  MinDistance -= BondThreshold;
};

void BondGraph::getMinMaxDistance(BondedParticle * const Walker, BondedParticle * const OtherWalker, double &MinDistance, double &MaxDistance, bool IsAngstroem)
{
  if (BondLengthMatrix == NULL) {// safety measure if no matrix has been parsed yet
    LOG(2, "INFO: Using Covalent radii criterion for [min,max] distances.");
    CovalentMinMaxDistance(Walker, OtherWalker, MinDistance, MaxDistance, IsAngstroem);
  } else {
    LOG(2, "INFO: Using Covalent radii criterion for [min,max] distances.");
    BondLengthMatrixMinMaxDistance(Walker, OtherWalker, MinDistance, MaxDistance, IsAngstroem);
  }
}
